@@ -61,8 +61,10 @@ void tools::Cygwin::Linker::AddLibGCC(const ArgList &Args,
   ToolChain::RuntimeLibType RLT = getToolChain().GetRuntimeLibType(Args);
   bool Static = Args.hasArg(options::OPT_static_libgcc) ||
                 Args.hasArg(options::OPT_static);
-
+  bool Shared = Args.hasArg(options::OPT_shared);
+  bool CXX = getToolChain().getDriver().CCCIsCXX();
   if (RLT == ToolChain::RLT_Libgcc) {
+
     if (Static) {
       CmdArgs.push_back("-lgcc");
       CmdArgs.push_back("-lgcc_eh");
@@ -72,10 +74,7 @@ void tools::Cygwin::Linker::AddLibGCC(const ArgList &Args,
     }
   } else {
     AddRunTimeLibs(getToolChain(), getToolChain().getDriver(), CmdArgs, Args);
-    if (Static)
-      CmdArgs.push_back("-lgcc_eh");
-    else
-      CmdArgs.push_back("-lgcc_s");
+    CmdArgs.push_back("-lunwind");
   }
 
   CmdArgs.push_back("-lcygwin");
@@ -100,14 +99,6 @@ void tools::Cygwin::Linker::ConstructJob(Compilation &C, const JobAction &JA,
   // handled somewhere else.
   Args.ClaimAllArgs(options::OPT_w);
 
-  StringRef LinkerName = Args.getLastArgValue(options::OPT_fuse_ld_EQ, "ld");
-  if (LinkerName.equals_lower("lld")) {
-    CmdArgs.push_back("-flavor");
-    CmdArgs.push_back("gnu");
-  } else if (!LinkerName.equals_lower("ld")) {
-    D.Diag(diag::err_drv_unsupported_linker) << LinkerName;
-  }
-
   if (!D.SysRoot.empty())
     CmdArgs.push_back(Args.MakeArgString("--sysroot=" + D.SysRoot));
 
@@ -115,38 +106,46 @@ void tools::Cygwin::Linker::ConstructJob(Compilation &C, const JobAction &JA,
     CmdArgs.push_back("-s");
 
   CmdArgs.push_back("-m");
-  if (TC.getArch() == llvm::Triple::x86)
+  switch (TC.getArch()) {
+  case llvm::Triple::x86:
     CmdArgs.push_back("i386pe");
-  if (TC.getArch() == llvm::Triple::x86_64)
+    break;
+  case llvm::Triple::x86_64:
     CmdArgs.push_back("i386pep");
+    break;
+  default:
+    llvm_unreachable("Unsupported target architecture.");
+  }
 
-  if (TC.getArch() == llvm::Triple::x86) {
-    CmdArgs.push_back("--wrap");
-    CmdArgs.push_back("_Znwj");
-    CmdArgs.push_back("--wrap");
-    CmdArgs.push_back("_Znaj");
-    CmdArgs.push_back("--wrap");
-    CmdArgs.push_back("_ZnwjRKSt9nothrow_t");
-    CmdArgs.push_back("--wrap");
-    CmdArgs.push_back("_ZnajRKSt9nothrow_t");
-  } else {
-    CmdArgs.push_back("--wrap");
-    CmdArgs.push_back("_Znwm");
-    CmdArgs.push_back("--wrap");
-    CmdArgs.push_back("_Znam");
-    CmdArgs.push_back("--wrap");
-    CmdArgs.push_back("_ZnwmRKSt9nothrow_t");
-    CmdArgs.push_back("--wrap");
-    CmdArgs.push_back("_ZnamRKSt9nothrow_t");
+  if (!Args.getLastArgValue(options::OPT_fuse_ld_EQ, "link").equals_lower("lld")) {
+    if (TC.getArch() == llvm::Triple::x86) {
+      CmdArgs.push_back("--wrap");
+      CmdArgs.push_back("_Znwj");
+      CmdArgs.push_back("--wrap");
+      CmdArgs.push_back("_Znaj");
+      CmdArgs.push_back("--wrap");
+      CmdArgs.push_back("_ZnwjRKSt9nothrow_t");
+      CmdArgs.push_back("--wrap");
+      CmdArgs.push_back("_ZnajRKSt9nothrow_t");
+    } else {
+      CmdArgs.push_back("--wrap");
+      CmdArgs.push_back("_Znwm");
+      CmdArgs.push_back("--wrap");
+      CmdArgs.push_back("_Znam");
+      CmdArgs.push_back("--wrap");
+      CmdArgs.push_back("_ZnwmRKSt9nothrow_t");
+      CmdArgs.push_back("--wrap");
+      CmdArgs.push_back("_ZnamRKSt9nothrow_t");
+    }
+    CmdArgs.push_back("--wrap");
+    CmdArgs.push_back("_ZdlPv");
+    CmdArgs.push_back("--wrap");
+    CmdArgs.push_back("_ZdaPv");
+    CmdArgs.push_back("--wrap");
+    CmdArgs.push_back("_ZdlPvRKSt9nothrow_t");
+    CmdArgs.push_back("--wrap");
+    CmdArgs.push_back("_ZdaPvKSt9nothrow_t");
   }
-  CmdArgs.push_back("--wrap");
-  CmdArgs.push_back("_ZdlPv");
-  CmdArgs.push_back("--wrap");
-  CmdArgs.push_back("_ZdaPv");
-  CmdArgs.push_back("--wrap");
-  CmdArgs.push_back("_ZdlPvRKSt9nothrow_t");
-  CmdArgs.push_back("--wrap");
-  CmdArgs.push_back("_ZdaPvKSt9nothrow_t");
 
   if (Args.hasArg(options::OPT_mwindows)) {
     CmdArgs.push_back("--subsystem");
@@ -163,7 +162,7 @@ void tools::Cygwin::Linker::ConstructJob(Compilation &C, const JobAction &JA,
       CmdArgs.push_back("--dll");
     else if (Args.hasArg(options::OPT_shared))
       CmdArgs.push_back("--shared");
-    if (Args.hasArg(options::OPT_rdynamic))
+    else if (Args.hasArg(options::OPT_rdynamic))
       CmdArgs.push_back("--export-all-symbols");
     CmdArgs.push_back("-Bdynamic");
     if (Args.hasArg(options::OPT_mdll) || Args.hasArg(options::OPT_shared)) {
@@ -192,9 +191,9 @@ void tools::Cygwin::Linker::ConstructJob(Compilation &C, const JobAction &JA,
   Args.AddLastArg(CmdArgs, options::OPT_Z_Flag);
 
   if (!Args.hasArg(options::OPT_nostdlib, options::OPT_nostartfiles)) {
-    if (Args.hasArg(options::OPT_shared))
+    if (Args.hasArg(options::OPT_shared) || Args.hasArg(options::OPT_mdll)) {
       CmdArgs.push_back(Args.MakeArgString(TC.GetFilePath("crtbeginS.o")));
-    else {
+    } else {
       CmdArgs.push_back(Args.MakeArgString(TC.GetFilePath("crt0.o")));
       if (Args.hasArg(options::OPT_pg))
         CmdArgs.push_back(Args.MakeArgString(TC.GetFilePath("gcrt0.o")));
@@ -210,8 +209,7 @@ void tools::Cygwin::Linker::ConstructJob(Compilation &C, const JobAction &JA,
 
   // TODO: Add profile stuff here
 
-  if (D.CCCIsCXX() &&
-      !Args.hasArg(options::OPT_nostdlib, options::OPT_nodefaultlibs)) {
+  if (TC.ShouldLinkCXXStdlib(Args)) {
     bool OnlyLibstdcxxStatic = Args.hasArg(options::OPT_static_libstdcxx) &&
                                !Args.hasArg(options::OPT_static);
     if (OnlyLibstdcxxStatic)
@@ -234,7 +232,7 @@ void tools::Cygwin::Linker::ConstructJob(Compilation &C, const JobAction &JA,
       if (Args.hasArg(options::OPT_pg))
         CmdArgs.push_back("-lgmon");
 
-      // add system libraries
+      // Add system libraries.
       if (Args.hasArg(options::OPT_mwindows)) {
         CmdArgs.push_back("-lgdi32");
         CmdArgs.push_back("-lcomdlg32");
@@ -246,8 +244,6 @@ void tools::Cygwin::Linker::ConstructJob(Compilation &C, const JobAction &JA,
 
       if (Args.hasArg(options::OPT_static))
         CmdArgs.push_back("--end-group");
-      else if (!LinkerName.equals_lower("lld"))
-        AddLibGCC(Args, CmdArgs);
     }
 
     if (!Args.hasArg(options::OPT_nostartfiles)) {
@@ -259,7 +255,7 @@ void tools::Cygwin::Linker::ConstructJob(Compilation &C, const JobAction &JA,
       CmdArgs.push_back(Args.MakeArgString(TC.GetFilePath("crtend.o")));
     }
   }
-  const char *Exec = Args.MakeArgString(TC.GetProgramPath(LinkerName.data()));
+  const char *Exec = Args.MakeArgString(TC.GetLinkerPath());
   C.addCommand(llvm::make_unique<Command>(JA, *this, Exec, CmdArgs, Inputs));
 }
 
@@ -288,7 +284,8 @@ void toolchains::Cygwin::findGccLibDir() {
   Archs.emplace_back(getTriple().getArchName());
   Archs[0] += "-pc-cygwin";
   Archs.emplace_back("cygwin");
-  Arch = Archs[0].str();
+  if (Arch.empty())
+    Arch = Archs[0].str();
   // lib: Arch Linux, Ubuntu, Windows
   // lib64: openSUSE Linux
   for (StringRef CandidateLib : {"lib", "lib64"}) {
@@ -315,12 +312,33 @@ llvm::ErrorOr<std::string> toolchains::Cygwin::findGcc() {
   return make_error_code(std::errc::no_such_file_or_directory);
 }
 
+llvm::ErrorOr<std::string> toolchains::Cygwin::findClangRelativeSysroot() {
+  llvm::SmallVector<llvm::SmallString<32>, 2> Subdirs;
+  Subdirs.emplace_back(getTriple().str());
+  Subdirs.emplace_back(getTriple().getArchName());
+  Subdirs[1] += "-pc-cygwin";
+  StringRef ClangRoot =
+      llvm::sys::path::parent_path(getDriver().getInstalledDir());
+  StringRef Sep = llvm::sys::path::get_separator();
+  for (StringRef CandidateSubdir : Subdirs) {
+    if (llvm::sys::fs::is_directory(ClangRoot + Sep + CandidateSubdir)) {
+      Arch = CandidateSubdir;
+      return (ClangRoot + Sep + CandidateSubdir).str();
+    }
+  }
+  return make_error_code(std::errc::no_such_file_or_directory);
+}
+
 toolchains::Cygwin::Cygwin(const Driver &D, const llvm::Triple &Triple,
                          const ArgList &Args)
     : ToolChain(D, Triple, Args), CudaInstallation(D, Triple, Args) {
   getProgramPaths().push_back(getDriver().getInstalledDir());
 
-  if (llvm::ErrorOr<std::string> GPPName = findGcc())
+  // Look for <clang-bin>/../<triplet>; if found, use <clang-bin>/.. as the
+  // base as it could still be a base for a gcc setup with libgcc.
+  if (llvm::ErrorOr<std::string> TargetSubdir = findClangRelativeSysroot())
+    Base = llvm::sys::path::parent_path(TargetSubdir.get());
+  else if (llvm::ErrorOr<std::string> GPPName = findGcc())
     Base = llvm::sys::path::parent_path(
         llvm::sys::path::parent_path(GPPName.get()));
   else
@@ -334,6 +352,7 @@ toolchains::Cygwin::Cygwin(const Driver &D, const llvm::Triple &Triple,
   getFilePaths().push_back(
       (Base + Arch + llvm::sys::path::get_separator() + "lib").str());
   getFilePaths().push_back(getDriver().SysRoot + "/usr/lib");
+  getFilePaths().push_back(getDriver().SysRoot + "/usr/lib/w32api");
 }
 
 bool toolchains::Cygwin::IsIntegratedAssemblerDefault() const { return true; }
@@ -375,8 +394,11 @@ bool toolchains::Cygwin::isPICDefaultForced() const {
   return getArch() == llvm::Triple::x86_64;
 }
 
-bool toolchains::Cygwin::UseSEHExceptions() const {
-  return getArch() == llvm::Triple::x86_64;
+llvm::ExceptionHandling
+toolchains::Cygwin::GetExceptionModel(const ArgList &Args) const {
+  if (getArch() == llvm::Triple::x86_64)
+    return llvm::ExceptionHandling::WinEH;
+  return llvm::ExceptionHandling::DwarfCFI;
 }
 
 void toolchains::Cygwin::AddCudaIncludeArgs(const ArgList &DriverArgs,
@@ -393,6 +415,12 @@ void toolchains::Cygwin::AddClangSystemIncludeArgs(const ArgList &DriverArgs,
   if (DriverArgs.hasArg(options::OPT_nostdinc))
     return;
 
+  if (!DriverArgs.hasArg(options::OPT_nobuiltininc)) {
+    SmallString<1024> P(getDriver().ResourceDir);
+    llvm::sys::path::append(P, "include");
+    addSystemInclude(DriverArgs, CC1Args, P.str());
+  }
+
   if (!DriverArgs.hasArg(options::OPT_nostdlibinc))
     addSystemInclude(DriverArgs, CC1Args, getDriver().SysRoot + "/usr/local/include");
 
@@ -417,11 +445,12 @@ void toolchains::Cygwin::AddClangCXXStdlibIncludeArgs(
       DriverArgs.hasArg(options::OPT_nostdincxx))
     return;
 
+  StringRef Slash = llvm::sys::path::get_separator();
+
   switch (GetCXXStdlibType(DriverArgs)) {
   case ToolChain::CST_Libcxx:
     addSystemInclude(DriverArgs, CC1Args,
-                     Base + "include" + llvm::sys::path::get_separator() +
-                         "c++" + llvm::sys::path::get_separator() + "v1");
+                     Base + "include" + Slash + "c++" + Slash + "v1");
     break;
 
   case ToolChain::CST_Libstdcxx:
@@ -436,7 +465,7 @@ void toolchains::Cygwin::AddClangCXXStdlibIncludeArgs(
     llvm::sys::path::append(CppIncludeBases[3], "include", "c++");
     for (auto &CppIncludeBase : CppIncludeBases) {
       addSystemInclude(DriverArgs, CC1Args, CppIncludeBase);
-      CppIncludeBase += llvm::sys::path::get_separator();
+      CppIncludeBase += Slash;
       addSystemInclude(DriverArgs, CC1Args, CppIncludeBase + Arch);
       addSystemInclude(DriverArgs, CC1Args, CppIncludeBase + "backward");
     }
